import java.io.*;
import java.util.*;

class Record { // 
  AllThread partner;
  char   operation;
  int    index, data;
  Record( int data ) {  this.data = data; }
  Record( AllThread th, char op, int index, int data ) {
    partner = th; operation = op;
    this.index = index;
    this.data = data;
  }
}

class DaemonThread extends Thread { // A szerver szl osztlya

  private RandomAccessFile file, // Adatfjl
                           logfile, // Logfjl
                           message; // zenetfjl
  private Vector v = new Vector(200), // krsek vektora
                 partners = new Vector(10); // partnerek vektora

  private int first, // els kiszolgland krs vektorindexe
              last; // utols kiszolgland krs vektorindexe + 1

  DaemonThread(String filename) {
    super( "Szerver" );
    first = last = 0; // kezdetben nincs kiszolgland krs
    this.setDaemon(true); // dmonknt fut a szl
    try { // fjlok megnyitsa
      file = new RandomAccessFile( filename, "rw" );
    } catch (Exception e) {System.out.println("Fjl hiba:"+e);}
    try { logfile = new RandomAccessFile("DaemonTest.log", "rw"); 
          message = new RandomAccessFile("DaemonTest.msg", "rw"); 
    }  catch (IOException e) {System.out.println("Fjl hiba:"+e);}
    System.out.println( "A "+getName()+" szl" + 
                        (isDaemon() ? " " : " nem ") + "dmonknt fut." );
  }
  public void run() { // krsek kiszolglsa
    try { sleep(500); } catch (InterruptedException e) {}
    while ( partners.size() > 0) { // vannak kiszolgland szlak?
      try { sleep(700); } catch (InterruptedException e) {}
      System.out.println( "A "+getName()+ " szl " + partners.size() +
                          " partnert szolgl ki.\tKrsek vektorindexei: " +
                          first + " - " + (last-1) );
      if (first != last) // ha van kiszolgland krs,
        theWork(); // akkor azt kiszolgljuk
      yield();     // engedjk a tbbieket is dolgozni
    }
    while (first != last) { // megmaradt krsek kiszolglsa
      System.out.println( "\tMaradk krsek kiszolglsa a " +
                          first + " - " + (last-1) + " vektorindexekrl");
      theWork();
    }
  }
  private void theWork() { // soronkvetkez krs kiszolglsa
    Record rec;            // segdvlt. az akt. zenethez
    boolean alive_thread;  // fut-e mg az zenet kldje

    rec = (Record) v.elementAt( first );  // a sor els eleme
    first = (first + 1) % 200;
    try { file.seek( rec.index * 4 ); } catch (IOException e) {};
          // 4 byte-os integer szmok, ezrt szorozzuk
    alive_thread = partners.contains( rec.partner );
    if (rec.operation  == 'r') {  // ha olvasni akar
      try { rec.data = file.readInt(); } catch (IOException e) {};
      if ( alive_thread ) { // ha mg fut, akkor visszaadjuk az eredmnyt
        rec.partner.putRecord( rec.data );
        rec.partner.resume();   // s tovbb engedjk
      }
    } else {  // ha rni akart
      try { file.writeInt( rec.data ); } catch (IOException e) {};
    }
    try { 
      if ( alive_thread ) { // logfile-ba, hogy mit rt
        logfile.writeChars( rec.partner.getName() + ":" + rec.operation +
                            " " + rec.index + ">" + rec.data + "\n" );
      } else
        logfile.writeChars( "HALTED thread:" + rec.operation +
                            " " + rec.index + ">" + rec.data + "\n" );
    } catch (IOException e) {}
  }
  public void sendMessage( Record rec ) { // zenetkld elj.
    v.insertElementAt(rec, last); // berakja a sor vgire
    if( !partners.contains(rec.partner) ) { // ha ez a partner mg nem
      partners.addElement(rec.partner); // ismert, akkor felveszi a listba
    }
    try { // ki s milyen zenetet kldtt
      message.writeChars( rec.partner.getName() + ":" + rec.operation +
                          " " + rec.index + 
                          (rec.operation=='w' ? " > " + rec.data : "") +
                          "\n" );
    } catch (IOException e) {}
    last = (last + 1) % 200;
    if (rec.operation == 'r')  rec.partner.suspend(); // ha olvas, akkor vrjon
  }  
  public void haltMessage( Thread th ) { // lell a partner
    partners.removeElement( th );
    System. out.println( th.getName() + " is halted..." );
  }
}

class AllThread extends Thread {  // a kliensek osztlya
  static private int counter = 0; // hny db van sszesen
  DaemonThread server;            // a szerver hivatkozsa
  int data;                       // segdvlt.

  AllThread( DaemonThread server ) {
    super( "Work # " + (++counter) );  // elnevezs s
    this.server = server;              // a szerver hivatkozs
  }
  public void run() {
    Random rnd = new Random();  // vletlenszmokhoz
    int index = 10, dat;        // csak a 10-es index rdekes

    try {
    while (true) {
      try { // itt lehetne a bels munka a kliensben
        sleep( 100 ); //* (1 + rnd.nextInt() % 3) ); 
      } catch (InterruptedException e) {}
      if (rnd.nextInt() % 5 > 0) { // nagyobb esllyel r
        server.sendMessage( new Record( this, 'w', index, 
                            (dat = (rnd.nextInt() % 100)) ));
        System.out.println( getName() + " has written(" + index + "):" + dat );
      } else {                     // mint olvas
        server.sendMessage( new Record( this, 'r', index, 0 ));
        System.out.println( getName() + " has read(" + index + "):" + data );
      }
    }
    } catch (ThreadDeath de) {    // ha lelltjk a szlat
      server.haltMessage( this ); // kld egy zenetet
      throw(de);
    }
  }
  public void putRecord( int data ) { // ezen kldi vissza a szerver
    this.data = data;                 // az olvasott rtket
  }
}

public class DaemonTest {
  static public void main(String s[]) {
    DaemonThread server;  // van egy szerver
    AllThread a, b, c;    // s 3 kliens

    server = new DaemonThread("DaemonTest.bin");
    System.out.println("Start the threads");
    a = new AllThread( server ); a.start(); 
    b = new AllThread( server ); b.start(); 
    c = new AllThread( server ); c.start(); 
    server.start();

    try { Thread.sleep(12000); } catch (InterruptedException ex) {}
    System.out.println("Stop the regular threads");
    a.stop(); 
    b.stop(); 
    c.stop(); 
    try { Thread.sleep(2000); } catch (InterruptedException ex) {}
  }
}
